# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


IDP_ENTITY_ID = 'https://localhost/v3/OS-FEDERATION/saml2/idp'
IDP_SSO_ENDPOINT = 'https://localhost/v3/OS-FEDERATION/saml2/SSO'

# Organization info
IDP_ORGANIZATION_NAME = 'ACME INC'
IDP_ORGANIZATION_DISPLAY_NAME = 'ACME'
IDP_ORGANIZATION_URL = 'https://acme.example.com'

# Contact info
IDP_CONTACT_COMPANY = 'ACME Sub'
IDP_CONTACT_GIVEN_NAME = 'Joe'
IDP_CONTACT_SURNAME = 'Hacker'
IDP_CONTACT_EMAIL = 'joe@acme.example.com'
IDP_CONTACT_TELEPHONE_NUMBER = '1234567890'
IDP_CONTACT_TYPE = 'technical'
